#! /bin/env python2

from casadi import *
import math
from xml.etree import ElementTree as et

class Component(object):
	def __init__(self):
		self.var = SX()
		self.par = SX()
		self.con = SX()
		self.obj = SX(0)
		self.vlb = []
		self.vub = []
		self.clb = []
		self.cub = []
		self.v0 = []

	def add_to(self, other):
		self.var.append(other.var)
		self.par.append(other.par)
		self.con.append(other.con)
		self.obj += other.obj
		self.vlb += other.vlb
		self.vub += other.vub
		self.clb += other.clb
		self.cub += other.cub
		self.v0 += other.v0

class Terminal(Component):
	def __init__(self):
		Component.__init__(self)
		self.p = SX.sym('p')
		self.q = SX.sym('q')
		self.v = SX.sym('v')
		self.th = SX.sym('th')

		self.var.append(self.p)
		self.var.append(self.q)
		self.var.append(self.v)
		self.var.append(self.th)

		self.vlb.append(-2e19)
		self.vlb.append(-2e19)
		self.vlb.append(-2e19)
		#self.vlb.append(.9)
		self.vlb.append(-2e19)

		self.vub.append(2e19)
		self.vub.append(2e19)
		self.vub.append(2e19)
		#self.vub.append(1.1)
		self.vub.append(2e19)

		self.v0.append(0.)
		self.v0.append(0.)
		self.v0.append(1.)
		self.v0.append(0.)

class Line(Component):
	def __init__(self, g, b):
		Component.__init__(self)
		t1 = Terminal()
		t2 = Terminal()
		self.add_to(t1)
		self.add_to(t2)
		self.t1 = t1
		self.t2 = t2

		p1 = t1.p
		q1 = t1.q
		p2 = t2.p
		q2 = t2.q
		vv1 = t1.v*t1.v
		vv2 = t2.v*t2.v
		vv12 = t1.v*t2.v
		s12 = sin(t1.th - t2.th)
		c12 = cos(t1.th - t2.th)

		self.con.append(-p1 + g*vv1 - g*vv12*c12 - b*vv12*s12);
		self.con.append(-p2 + g*vv2 - g*vv12*c12 + b*vv12*s12);
		self.con.append(-q1 - b*vv1 + b*vv12*c12 - g*vv12*s12);
		self.con.append(-q2 - b*vv2 + b*vv12*c12 + g*vv12*s12);

		for i in range(4):
			self.clb.append(0.)
			self.cub.append(0.)

class Experiment(Component):
	def __init__(self, ifn, ofn):
		Component.__init__(self)
		rh = 0.07 # resistance of house conductors
		xh = 0.07 # reactance of house conductors
		rp = 0.1 # resistance of shared conductor
		xp = 0.1 # reactance of shared conductor

		gh = rh/(rh**2 + xh**2) # conductance
		bh = -xh/(rh**2 + xh**2) # susceptance
		gp = rp/(rp**2 + xp**2) # conductance
		bp = -xp/(rp**2 + xp**2) # susceptance

		ph1 = .5
		qh1 = .1
		ph2 = .2
		qh2 = .1

		l1 = Line(gh, bh)
		self.add_to(l1)
		l2 = Line(gh, bh)
		self.add_to(l2)
		l3 = Line(gp, bp)
		self.add_to(l3)

		# Creating parameters
		pinv = SX.sym('pinv')
		qinv = SX.sym('qinv')
		v2set = SX.sym('v2set')

		self.par.append(pinv)
		self.par.append(qinv)
		self.par.append(v2set)

		# Creating constraints and bounds
		self.con.append(l1.t1.p + ph1 - pinv)
		self.con.append(l1.t1.q + qh1 - qinv)
		self.con.append(l2.t1.p + ph2)
		self.con.append(l2.t1.q + qh2)

		self.con.append(l1.t2.p + l2.t2.p + l3.t1.p)
		self.con.append(l1.t2.q + l2.t2.q + l3.t1.q)
		self.con.append(l1.t2.v - l3.t1.v)
		self.con.append(l2.t2.v - l3.t1.v)
		self.con.append(l1.t2.th - l3.t1.th)
		self.con.append(l2.t2.th - l3.t1.th)

		self.con.append(l3.t2.v - v2set)
		self.con.append(l3.t2.th)

		for i in range(12):
			self.clb.append(0.)
			self.cub.append(0.)

		self.ofn = ofn

		svgin = open(ifn).read()
		svgin = svgin.replace('xmlns=', 'xmlnamespace=') # work around library
		root = et.fromstring(svgin)
		root.attrib['xmlns'] = root.attrib['xmlnamespace']
		root.attrib.pop('xmlnamespace')
		texts = {}
		for txt in root.findall('.//text'):
			texts[txt.attrib['id']] = txt.find('tspan')
		self.texts = texts
		self.root = root

	def prepare(self):
		# Building solver
		nlp = SXFunction(nlpIn(x=self.var, p=self.par),
				nlpOut(f=self.obj, g=self.con))
		nlp.init()
		solver = IpoptSolver(nlp)
		solver.setOption('print_level', 0);
		solver.setOption('print_time', False);
		solver.init()
		solver.setInput(self.vlb, 'lbx')
		solver.setInput(self.vub, 'ubx')
		solver.setInput(self.v0, 'x0')
		solver.setInput(self.clb, 'lbg')
		solver.setInput(self.cub, 'ubg')
		self.solver = solver
	
	def solve(self, pval):
		self.solver.setInput(pval, 'p')
		self.solver.solve()
		stat = self.solver.getStat('return_status')
		print(stat)
		if stat == 'Solve_Succeeded':
			return True
		else:
			return False
	
	def update_svg(self):
		# Retrieving data
		varv = ex.solver.getOutput('x')

		self.texts['l1v1'].text = '%.2f' % varv[2]
		self.texts['l2v1'].text = '%.2f' % varv[8+2]
		self.texts['l3v1'].text = '%.2f' % varv[16+2]
		self.texts['l3v2'].text = '%.2f' % varv[16+6]
		self.texts['l1p1'].text = '%+.2f' % varv[0]
		self.texts['l2p1'].text = '%+.2f' % varv[8]
		self.texts['l3p1'].text = '%+.2f' % varv[16]
		self.texts['l3p2'].text = '%+.2f' % -varv[16+4]
		self.texts['l1q1'].text = '%+.2f' % varv[1]
		self.texts['l2q1'].text = '%+.2f' % varv[8+1]
		self.texts['l3q1'].text = '%+.2f' % varv[16+1]
		self.texts['l3q2'].text = '%+.2f' % -varv[16+5]

		f = open(self.ofn, 'w')
		f.write('<?xml version=\"1.0\" encoding="UTF-8" standalone=\"no\"?>\n')
		f.write(et.tostring(self.root))
		f.close()

	def run(self, p, q):
		pval = []
		pval.append(p)
		pval.append(q)
		pval.append(1.1)
		if self.solve(pval):
			self.update_svg()


ex = Experiment('image.svg', 'out.svg')
ex.prepare()

#tap ratio typically static and set to handle full range of peaks and troughs
#ex.run(p=0, q=0)
#ex.run(p=1, q=0)
#ex.run(p=1.5, q=0)
#ex.run(p=1.5, q=-0.5)
#ex.run(p=1.5, q=-0.9)
#ex.run(p=0., q=0.5)
